/***********************************************************************/
/*                                                                     */
/* RIGI C Parser --- Copyright (C) 1996-1999                           */
/*                    The Rigi Group, University of Victoria           */
/*                                                                     */
/* see the file license.terms for usage and copying policies           */
/*                                                                     */
/* Maintainer: Johannes Martin (jmartin@csr.csc.uvic.ca)               */
/*                                                                     */
/***********************************************************************/
#include <stdlib.h>	// every C source file has it ...
#include <string.h>	// for strdup
#include <stdio.h>

#include "String.h"
#include "List.h"
#include "globals.h"
#include "clex.h"
#include "cinit.h"
#include "symtbl.h"
#include "parse.h"

static List<String> current_struct;     // Structure being processed
static List<String> current_func;       // Function being processed

static char *current_type = NULL;	// Type being processed
static char *current_functype = NULL;	// Type of function being processed
static ParseInfo current_typedeftype;   // src-type of typedef being processed
static int current_storage;             // current storage type (static/...)
static bool current_prototype;          // is the current type a prototype
                                        // definition?

void relation(const char *r, const char *s, const char *d,
              const ParseInfo& p, int writeloc)
{
  if (!r) {
    yyerror("internal error: first component of tuple is NULL");
    r = "NULL";
  }
  if (!s) {
    yyerror("internal error: second component of tuple is NULL");
    s = "NULL";
  }
  if (!d) {
    yyerror("internal error: third component of tuple is NULL");
    d = "NULL";
  }
  if (!quiet) {
    if (write4 && writeloc)
      fprintf(outfile, "%s\t%s\t%s\t%s,%d,%d\n", r, s, d,
	      p.File(), p.Line(), p.Col());
    else
      fprintf(outfile, "%s\t%s\t%s\n", r, s, d);
  }
}

void relation(const char *r, const char *s, int d,
              const ParseInfo& p, int writeloc)
{
  if (!r) {
    yyerror("internal error: first component of tuple is NULL");
    r = "NULL";
  }
  if (!s) {
    yyerror("internal error: second component of tuple is NULL");
    s = "NULL";
  }
  if (!quiet) {
    if (write4 && writeloc)
      fprintf(outfile, "%s\t%s\t%d\t%s,%d,%d\n", r, s, d,
	      p.File(), p.Line(), p.Col());
    else
      fprintf(outfile, "%s\t%s\t%d\n", r, s, d);
  }
}

void setprototype(bool b)
{
    current_prototype = b;
}

bool getprototype() 
{
    return current_prototype;
}

void settype(const char *s)
{
    if (current_type) free(current_type);
    current_type = s ? strdup(s) : NULL;
    if (symdebug && current_type)
        fprintf(stderr, "settype: setting type to %s\n", s);
}

const char *gettype()
{
    return current_type;
}

void setfunctype(const char *s)
{
    if (current_functype) free(current_functype);
    current_functype = s ? strdup(s) : NULL;
    if (symdebug && current_functype)
        fprintf(stderr, "setftype: setting type to %s\n", s);
}

const char *getfunctype()
{
    return current_functype;
}

void settypedeftype(const ParseInfo& info)
{
  current_typedeftype = info;
}

const ParseInfo& gettypedeftype()
{
  return current_typedeftype;
}

void setstorage(int storage)
{
  current_storage = storage;
}

int getstorage()
{
  return current_storage;
}

void setstruct(const char* s)
{
  current_struct.addTail(s);
  if (symdebug)
    fprintf(stderr, "setstruct: setting struct to %s\n", s);
}

const char *getstruct()
{
  return current_struct.empty() ? NULL : (const char*) current_struct.getTail();
}

void popstruct()
{
  current_struct.remTail();
}

void startfunc(const ParseInfo& p)
{
  current_func.addTail(p.QualifiedName());
}

const char *getfunc()
{
  return current_func.empty() ? "*" : (const char*) current_func.getTail();
}

bool infunc()
{
  return !current_func.empty();
}

void endfunc(const ParseInfo& p) {
  relation("type", p.QualifiedName(), "Function", p, write4);
  if (fileline) {
    relation("file", p.QualifiedName(), p.File(), p, 0);
    relation("lineno", p.QualifiedName(), p.Line(), p, 0);
  }
  if (fileNode)
    relation("isInFile", p.QualifiedName(), p.File(), p, 0);
  relation("returns", p.QualifiedName(), p.Type(), p, write4);
  current_func.remTail();
}

void endproto(const ParseInfo& p) {
  relation("type", p.QualifiedName(), "Prototype", p, write4);
  if (fileline) {
    relation("file", p.QualifiedName(), p.File(), p, 0);
    relation("lineno", p.QualifiedName(), p.Line(), p, 0);
  }
  if (fileNode) 
    relation("isInFile", p.QualifiedName(), p.File(), p, 0);
  relation("returns", p.QualifiedName(), p.Type(), p, write4);
  current_func.remTail();
  if (infunc())
    relation("isDefinedIn", p.QualifiedName(), getfunc(), p, write4);
}

void newvar(const ParseInfo& p)
{
  relation("type", p.QualifiedName(), 
	   p.Prototype() ? "Prototype" : "Variable", p, write4);
  if (fileline) {
    relation("file", p.QualifiedName(), p.File(), p, 0);
    relation("lineno", p.QualifiedName(), p.Line(), p, 0);
  }
  if (fileNode)
    relation("isInFile", p.QualifiedName(), p.File(), p, 0);
  relation("hasType", p.QualifiedName(), p.Type(), p, write4);
  if (infunc())
    relation("isDefinedIn", p.QualifiedName(), getfunc(), p, write4);
}

void newconst(const ParseInfo& p)
{
    relation("type", p.QualifiedName(), "Constant", p, write4);
    if (fileline) {
      relation("file", p.QualifiedName(), p.File(), p, 0);
      relation("lineno", p.QualifiedName(), p.Line(), p, 0);
    }
    if (fileNode) 
      relation("isInFile", p.QualifiedName(), p.File(), p, 0);
    relation("hasType", p.QualifiedName(), p.Type(), p, write4);
    if (infunc())
      relation("isDefinedIn", p.QualifiedName(), getfunc(), p, write4);
}

void newtypedef(const ParseInfo& p)
{
    relation("type", p.QualifiedName(), "Datatype", p, write4);
    if (fileline) {
      relation("file", p.QualifiedName(), p.File(), p, 0);
      relation("lineno", p.QualifiedName(), p.Line(), p, 0);
    }
    if (fileNode)
      relation("isInFile", p.QualifiedName(), p.File(), p, 0);
    if (infunc())
      relation("isDefinedIn", p.QualifiedName(), getfunc(), p, write4);
    if (p.Type())
      relation("isTheSameAs", p.QualifiedName(), p.Type(), p, write4);
}

void newcall(const ParseInfo& p)
{
    relation("calls", getfunc(), p.QualifiedName(), p, write4);
}

void newdata(const ParseInfo& p)
{
  if (!p.Type()) {
    yyerror("internal error: newdata - invalid data type");
    return;
  }
  relation("accesses", getfunc(), p.Type(), p, write4);
}

void newref(const ParseInfo& p)
{
    relation("references", getfunc(), p.QualifiedName(), p, write4);
}

void newprototypes(const ParseInfo& protoInfo, const ParseInfo& implInfo) {
  relation("prototypes", protoInfo.QualifiedName(),
	   implInfo.QualifiedName(), protoInfo, write4);
}

void newdeclares(const ParseInfo& externInfo, const ParseInfo& newInfo) {
  relation("declares", externInfo.QualifiedName(),
	   newInfo.QualifiedName(), externInfo, write4);
}

void newtagged(const ParseInfo& p)
{
    relation("tagged", p.QualifiedName(), p.QualifiedName(), p, write4);
}

void newfile(const char* fileName)
{
  if (fileNode && strcmp(fileName + (strlen(fileName) - 7), "<stdin>"))
    relation("type", fileName, "File", ParseInfo(), false);
}
